/* gtk-qn-x11-theme - A gtk+ theme engine.
   Copyright (C) 2000  Mark Slicker <jamess1@wwnet.com>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <math.h>
#include <gtk/gtk.h>

#define DETAIL(xx)   ((detail) && (!strcmp(xx, detail)))

/* internal functions */
static void         draw_hline(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x1,
			       gint x2,
			       gint y);
static void         draw_vline(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint y1,
			       gint y2,
			       gint x);
static void         draw_shadow(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height);

static void         draw_polygon(GtkStyle * style,
				 GdkWindow * window,
				 GtkStateType state_type,
				 GtkShadowType shadow_type,
				 GdkRectangle * area,
				 GtkWidget * widget,
				 gchar * detail,
				 GdkPoint * point,
				 gint npoints,
				 gint fill);
static void         draw_arrow(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GtkShadowType shadow_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       GtkArrowType arrow_type,
			       gint fill,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_diamond(GtkStyle * style,
				 GdkWindow * window,
				 GtkStateType state_type,
				 GtkShadowType shadow_type,
				 GdkRectangle * area,
				 GtkWidget * widget,
				 gchar * detail,
				 gint x,
				 gint y,
				 gint width,
				 gint height);
static void         draw_oval(GtkStyle * style,
			      GdkWindow * window,
			      GtkStateType state_type,
			      GtkShadowType shadow_type,
			      GdkRectangle * area,
			      GtkWidget * widget,
			      gchar * detail,
			      gint x,
			      gint y,
			      gint width,
			      gint height);
static void         draw_string(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				const gchar * string);
static void         draw_box(GtkStyle * style,
			     GdkWindow * window,
			     GtkStateType state_type,
			     GtkShadowType shadow_type,
			     GdkRectangle * area,
			     GtkWidget * widget,
			     gchar * detail,
			     gint x,
			     gint y,
			     gint width,
			     gint height);
static void         draw_flat_box(GtkStyle * style,
				  GdkWindow * window,
				  GtkStateType state_type,
				  GtkShadowType shadow_type,
				  GdkRectangle * area,
				  GtkWidget * widget,
				  gchar * detail,
				  gint x,
				  gint y,
				  gint width,
				  gint height);
static void         draw_check(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GtkShadowType shadow_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_option(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height);
static void         draw_cross(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GtkShadowType shadow_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_ramp(GtkStyle * style,
			      GdkWindow * window,
			      GtkStateType state_type,
			      GtkShadowType shadow_type,
			      GdkRectangle * area,
			      GtkWidget * widget,
			      gchar * detail,
			      GtkArrowType arrow_type,
			      gint x,
			      gint y,
			      gint width,
			      gint height);
static void         draw_tab(GtkStyle * style,
			     GdkWindow * window,
			     GtkStateType state_type,
			     GtkShadowType shadow_type,
			     GdkRectangle * area,
			     GtkWidget * widget,
			     gchar * detail,
			     gint x,
			     gint y,
			     gint width,
			     gint height);
static void         draw_shadow_gap(GtkStyle * style,
				    GdkWindow * window,
				    GtkStateType state_type,
				    GtkShadowType shadow_type,
				    GdkRectangle * area,
				    GtkWidget * widget,
				    gchar * detail,
				    gint x,
				    gint y,
				    gint width,
				    gint height,
				    GtkPositionType gap_side,
				    gint gap_x,
				    gint gap_width);
static void         draw_box_gap(GtkStyle * style,
				 GdkWindow * window,
				 GtkStateType state_type,
				 GtkShadowType shadow_type,
				 GdkRectangle * area,
				 GtkWidget * widget,
				 gchar * detail,
				 gint x,
				 gint y,
				 gint width,
				 gint height,
				 GtkPositionType gap_side,
				 gint gap_x,
				 gint gap_width);
static void         draw_extension(GtkStyle * style,
				   GdkWindow * window,
				   GtkStateType state_type,
				   GtkShadowType shadow_type,
				   GdkRectangle * area,
				   GtkWidget * widget,
				   gchar * detail,
				   gint x,
				   gint y,
				   gint width,
				   gint height,
				   GtkPositionType gap_side);
static void         draw_focus(GtkStyle * style,
			       GdkWindow * window,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_slider(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height,
				GtkOrientation orientation);
static void         draw_handle(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height,
				GtkOrientation orientation);

static void
draw_button(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height);

void
draw_trough (GtkStyle * style,
	     GdkWindow * window,
	     GdkRectangle * area,
	     int x, int y, int width, int height,
	     GdkColor *c);

void
draw_concave_box (GtkStyle * style,
		  GdkWindow * window,
		  GdkRectangle * area,
		  int x, int y, int width, int height,
		  GdkColor *c);

void
draw_beveled_box (GtkStyle * style,
		  GdkWindow * window,
		  GdkRectangle * area,
		  int x, int y, int width, int height,
		  GdkColor *c);

void
draw_vertical_gradient (GtkStyle * style,
			GdkWindow * window,
			GdkRectangle * area,
			int x, int y, int width, int height,
			GdkColor *a,
			GdkColor *b);

void
draw_cubic_vertical_gradient (GtkStyle * style,
			      GdkWindow * window,
			      GdkRectangle * area,
			      int x, int y, int width, int height,
			      GdkColor *base_color,
			      float a, float b, float c, float d);

#define clamp(value)				\
  (value > 0xffff)? 0xffff : value

#define mult(b, a, mul, cmap)					\
do {							\
  unsigned int value;					\
  value = (unsigned int)(a.red * mul);		\
  b.red = value > 0xffff ? 0xffff : value;		\
  value = (unsigned int)(a.green * mul);	\
  b.green = value > 0xffff ? 0xffff : value;		\
  value = (unsigned int)(a.blue * mul);		\
  b.blue = value > 0xffff ? 0xffff : value;		\
  gdk_color_alloc (cmap, &b);					\
} while (0)

/* internal data structs */

GtkStyleClass       flat_default_class =
{
  2,
  2,
  draw_hline,
  draw_vline,
  draw_shadow,
  draw_polygon,
  draw_arrow,
  draw_diamond,
  draw_oval,
  draw_string,
  draw_box,
  draw_flat_box,
  draw_check,
  draw_option,
  draw_cross,
  draw_ramp,
  draw_tab,
  draw_shadow_gap,
  draw_box_gap,
  draw_extension,
  draw_focus,
  draw_slider,
  draw_handle
};

static void
draw_hline(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x1,
	   gint x2,
	   gint y)
{
  GdkColor c;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, area);

  mult (c, style->bg[state_type], 0.285493827, style->colormap);
  gdk_gc_set_foreground (style->white_gc, &c);
  gdk_draw_line(window, style->white_gc, x1, y, x2 - 2, y);
  mult (c, style->bg[state_type], 1.111111111, style->colormap);
  gdk_gc_set_foreground (style->white_gc, &c);
  gdk_draw_line(window, style->white_gc, x1+1, y + 1, x2 - 1, y + 1);

  gdk_gc_set_foreground (style->white_gc, &style->white);

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, NULL);
}
/**************************************************************************/
static void
draw_vline(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint y1,
	   gint y2,
	   gint x)
{

  GdkColor c;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, area);

  mult (c, style->bg[state_type], 0.285493827, style->colormap);
  gdk_gc_set_foreground (style->white_gc, &c);
  gdk_draw_line(window, style->white_gc, x, y1, x, y2-2);
  mult (c, style->bg[state_type], 1.111111111, style->colormap);
  gdk_gc_set_foreground (style->white_gc, &c);
  gdk_draw_line(window, style->white_gc, x+1, y1+1, x+1, y2-1);

  gdk_gc_set_foreground (style->white_gc, &style->white);

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, NULL);
}
/**************************************************************************/
static void
draw_shadow(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height)
{
  GdkColor c;

  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
  
  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  if (area)
    gdk_gc_set_clip_rectangle (style->white_gc, area);
  
  switch (shadow_type)
    {
    case GTK_SHADOW_NONE:
      break;
      
    case GTK_SHADOW_IN:
      /* draw outer box */
      mult (c, style->bg[state_type], 0.699074074, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_line (window, style->white_gc,
                     x, y, x + width - 2, y);
      gdk_draw_line (window, style->white_gc,
                     x, y+1, x, y + height - 2);
      mult (c, style->bg[state_type], 1.111111111, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_line (window, style->white_gc,
                     x + 1, y + height - 1, x + width - 1, y + height - 1);
      gdk_draw_line (window, style->white_gc,
                     x + width - 1, y + 1, x + width - 1, y + height - 2);
      gdk_draw_point (window, style->bg_gc[state_type], x + width - 1, y);
      mult (c, style->bg[state_type], .787037037, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_point (window, style->white_gc, x, y + height - 1);
      /* draw inner box */
      mult (c, style->bg[state_type], 0.285493827, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_rectangle (window, style->white_gc, FALSE,
			  x+1, y+1, width-3, height-3);
      break;
      
    case GTK_SHADOW_OUT:
      /* draw outer box */
      mult (c, style->bg[state_type], 0.285493827, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_rectangle (window, style->white_gc, FALSE,
			  x, y, width-1, height-1);
      /* draw inner box */
      mult (c, style->bg[state_type], 1.111111111, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_line (window, style->white_gc,
                     x+1, y+1, x + width - 3, y+1);
      gdk_draw_line (window, style->white_gc,
                     x+1, y+2, x+1, y + height - 3);
      mult (c, style->bg[state_type], 0.699074074, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_line (window, style->white_gc,
                     x + 2, y + height - 2, x + width - 2, y + height - 2);
      gdk_draw_line (window, style->white_gc,
                     x + width - 2, y + 2, x + width - 2, y + height - 3);
      gdk_draw_point (window, style->bg_gc[state_type], x + width - 2, y+1);
      mult (c, style->bg[state_type], .787037037, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_point (window, style->white_gc, x+1, y + height - 2);
      break;
      
    case GTK_SHADOW_ETCHED_IN:
      mult (c, style->bg[state_type], 1.111111111, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_rectangle (window, style->white_gc, FALSE,
			  x+1, y+1, width-2, height-2);
      mult (c, style->bg[state_type], 0.285493827, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_rectangle (window, style->white_gc, FALSE,
			  x, y, width-2, height-2);

      break;
    case GTK_SHADOW_ETCHED_OUT:
      mult (c, style->bg[state_type], 0.285493827, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_rectangle (window, style->white_gc, FALSE,
			  x+1, y+1, width-2, height-2);
      mult (c, style->bg[state_type], 1.111111111, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_rectangle (window, style->white_gc, FALSE,
			  x, y, width-2, height-2);
      break;
    }

  gdk_gc_set_foreground (style->white_gc, &style->white);

  if (area)
    gdk_gc_set_clip_rectangle (style->white_gc, NULL);
}
/**************************************************************************/
static void
draw_polygon(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     GdkPoint * points,
	     gint npoints,
	     gint fill)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
  g_return_if_fail(points != NULL);


  if (area) {
    gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
    gdk_gc_set_clip_rectangle(style->fg_gc[state_type], area);
  }

  if (fill)
    gdk_draw_polygon(window, style->bg_gc[state_type], TRUE, points, npoints);
  gdk_draw_polygon(window, style->fg_gc[state_type], FALSE, points, npoints);

  if (area) {
    gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
    gdk_gc_set_clip_rectangle(style->fg_gc[state_type], NULL);
  }
}
/**************************************************************************/
static void
draw_arrow(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   GtkArrowType arrow_type,
	   gint fill,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  GdkGC              *gc;
  gint                half_width;
  gint                half_height;
  gint                xthik, ythik;
  GdkPoint            points[3];
  gchar               border = 1;
   int base, span, xoffset, yoffset;
   int i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1)) gdk_window_get_size(window, &width, &height);
  else if (width == -1)                gdk_window_get_size(window, &width, NULL);
  else if (height == -1)               gdk_window_get_size(window, NULL, &height);

  xthik = style->klass->xthickness;
  ythik = style->klass->ythickness;

  gc = style->fg_gc[state_type];

  if (DETAIL("menuitem") || DETAIL("arrow")) {
    border = 0;
  }


  if ((fill) && (border)) {
    draw_beveled_box (style, window, area, x, y, width, height,
		    &style->bg[state_type]);
  }

  if (DETAIL("spinbutton")) {
    draw_button (style, window, state_type, shadow_type, area, widget, detail, x, y, width, height);
  }

  if (area)
    gdk_gc_set_clip_rectangle(gc, area);


  switch (arrow_type) {
  case GTK_ARROW_UP:
    if (border)
      base = width / 2;
    else
      base    = width/2 +3;
    if (base%2 == 0) base++;
    xoffset = (width-base)/2;
    span = base/2 + 1;
    yoffset = (height+span)/2-1;
    for (i=0; i<span; i++) {
      gdk_draw_line(window, gc, x+xoffset+i, y+yoffset-i, 
		    x+xoffset+base-1-i, y+yoffset-i);
    }
    break;
  case GTK_ARROW_DOWN:
    if (border)
      base = width / 2;
    else
      base    = width/2 +3;
    if (base%2 == 0) base++;
    xoffset = (width-base)/2;
    span = base/2 + 1;
    yoffset = (height-span)/2;
    for (i=0; i<span; i++) {
      gdk_draw_line(window, gc, x+xoffset+i, y+yoffset+i, 
		    x+xoffset+base-1-i, y+yoffset+i);
    }
    break;
  case GTK_ARROW_RIGHT:
    if (GTK_CHECK_TYPE(widget, gtk_menu_item_get_type()))
      base = 7;
    else if (border)
      base = height / 2;
    else
      base = height/2 +3;
    if (base%2 == 0) base++;
    yoffset = (height-base)/2;
    span = base/2 + 1;
    xoffset = (width-span)/2;
    if (!border)
      xoffset++;

    for (i=0; i<span; i++) {
      gdk_draw_line(window, gc, x+xoffset+i, y+yoffset+i, 
		    x+xoffset+i, y+yoffset+base-1-i);
    }
    break;
  case GTK_ARROW_LEFT:
    if (border)
      base = height / 2;
    else
      base    = height/2 +3;
    if (base%2 == 0) base++;
    yoffset = (height-base)/2;
    span = base/2 + 1;
    xoffset = (width+span)/2-1;
    if (!border)
      xoffset--;
    for (i=0; i<span; i++) {
      gdk_draw_line(window, gc, x+xoffset-i, y+yoffset+i, 
		    x+xoffset-i, y+yoffset+base-1-i);
    }
    break;
  }

  if (area)
    gdk_gc_set_clip_rectangle(gc, NULL);
}
/**************************************************************************/
static void
draw_diamond(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     gint x,
	     gint y,
	     gint width,
	     gint height)
{
  gint                half_width;
  gint                half_height;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1)) gdk_window_get_size(window, &width, &height);
  else if (width == -1)                gdk_window_get_size(window, &width, NULL);
  else if (height == -1)               gdk_window_get_size(window, NULL, &height);

  half_width = width / 2;
  half_height = height / 2;

  if (area) {
    gdk_gc_set_clip_rectangle(style->fg_gc[state_type], area);
  }

  gdk_draw_line(window, style->fg_gc[state_type],
		x, y + half_height,
		x + half_width, y);
  gdk_draw_line(window, style->fg_gc[state_type],
	        x + half_width, y,
	        x + width - 1, y + half_height);
  gdk_draw_line(window, style->fg_gc[state_type],
         	x + half_width, y + height-1,
      	        x + width-1, y + half_height);
  gdk_draw_line(window, style->fg_gc[state_type],
		x, y + half_height,
		x + half_width, y + height-1);

  if (area) {
    gdk_gc_set_clip_rectangle(style->fg_gc[state_type], NULL);
  }
}
/**************************************************************************/
static void
draw_oval(GtkStyle * style,
	  GdkWindow * window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle * area,
	  GtkWidget * widget,
	  gchar * detail,
	  gint x,
	  gint y,
	  gint width,
	  gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
}
/**************************************************************************/
static void
draw_string(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    const gchar * string)
{
  GdkGC *fggc;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  fggc = style->fg_gc[state_type];

  if (area) {
    gdk_gc_set_clip_rectangle(fggc, area);
  }

  gdk_draw_string(window, style->font, fggc, x, y, string);

  if (area) {
    gdk_gc_set_clip_rectangle(fggc, NULL);
  }
}

/**************************************************************************/
static void
draw_box(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GtkShadowType shadow_type,
	 GdkRectangle * area,
	 GtkWidget * widget,
	 gchar * detail,
	 gint x,
	 gint y,
	 gint width,
	 gint height)
{
  GdkColor a, b;
  GdkPixmap *p;
  GtkStyle *parent_style;
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1)) gdk_window_get_size(window, &width, &height);
  else if (width == -1)                gdk_window_get_size(window, &width, NULL);
  else if (height == -1)               gdk_window_get_size(window, NULL, &height);

  if (DETAIL("button") || DETAIL("togglebutton") || DETAIL("buttondefault") || DETAIL ("optionmenu")) {
    draw_button(style, window, state_type, shadow_type,
		area, widget, detail, x, y, width, height);
  } else if (DETAIL ("slider")) {
    if (area)
      gdk_gc_set_clip_rectangle (style->white_gc, area);
    a.red = (unsigned short)(style->bg[GTK_STATE_NORMAL].red * .2361111);
    a.green = (unsigned short)(style->bg[GTK_STATE_NORMAL].green * .2361111);
    a.blue = (unsigned short)(style->bg[GTK_STATE_NORMAL].blue * .2361111);
    gdk_color_alloc (style->colormap, &a);
    gdk_gc_set_foreground (style->white_gc, &a);

    if (GTK_IS_HSCROLLBAR(widget)) {
      gdk_draw_line (window, style->white_gc, 
		     x, y, x, y+height-1);
      gdk_draw_line (window, style->white_gc, 
		     x+width-1, y, x+width-1, y+height-1);
      draw_beveled_box (style, window, area, x+1, y,
			width-2, height, &style->bg[state_type]);
    } else {
      gdk_draw_line (window, style->white_gc, 
		     x, y, x+width-1, y);
      gdk_draw_line (window, style->white_gc, 
		     x, y+height-1, x+width-1, y+height-1);
      draw_beveled_box (style, window, area, x, y+1,
			width, height-2, &style->bg[state_type]);
    }
    if (area)
      gdk_gc_set_clip_rectangle (style->white_gc, NULL);
  } else if (DETAIL ("trough")) {
    if (area)
      gdk_gc_set_clip_rectangle (style->white_gc, area);

    parent_style = gtk_widget_get_style(GTK_WIDGET(widget->parent));

    if (GTK_IS_SCROLLBAR (widget)) {
      p = gdk_pixmap_new (window, width, height, style->depth);

      gtk_paint_shadow (parent_style, p, GTK_STATE_NORMAL, shadow_type,
			area, widget, detail, x, y, width, height);

      a.red = (unsigned short)(style->bg[GTK_STATE_NORMAL].red * .2361111);
      a.green = (unsigned short)(style->bg[GTK_STATE_NORMAL].green * .2361111);
      a.blue = (unsigned short)(style->bg[GTK_STATE_NORMAL].blue * .2361111);
      gdk_color_alloc (style->colormap, &a);
      gdk_gc_set_foreground (style->white_gc, &a);

      if (GTK_IS_HSCROLLBAR (widget)) {
	gdk_draw_line (p, style->white_gc, 
		       x+height-2, y+2, x+height-2, y+height-3);
	gdk_draw_line (p, style->white_gc, 
		       x+width-height+1, y+2, x+width-height+1, y+height-3);
	draw_trough (style, p, area, x+height-1, y+2,
		     width-2*height+2, height-4, &style->bg[GTK_STATE_NORMAL]);
      } else {
	gdk_draw_line (p, style->white_gc, 
		       x+2, y+width-2, x+width-3, y+width-2);
	gdk_draw_line (p, style->white_gc, 
		       x+2, y+height-width+1, x+width-3, y+height-width+1);
	draw_trough (style, p, area, x+2, y+width-1,
		     width-4, height-2*width+2, &style->bg[GTK_STATE_NORMAL]);
      }

      if (GTK_WIDGET_NO_WINDOW (widget))
	gdk_draw_pixmap (window, style->white_gc, p, 0, 0, 0, 0, width, height);
      else
	gdk_window_set_back_pixmap (window, p, 0);
      gdk_pixmap_unref (p);
    
      if (area) {
	gdk_window_clear_area (window, area->x, area->y,
			     area->width, area->height);
      } else
	gdk_window_clear (window);
    } else {
      gtk_paint_shadow (parent_style, window, GTK_STATE_NORMAL, shadow_type,
			area, widget, detail, x, y, width, height);

      draw_trough (style, window, area, x+2, y+2,
		   width-4, height-4, &style->bg[GTK_STATE_NORMAL]);
    }

    if (area)
      gdk_gc_set_clip_rectangle (style->white_gc, NULL);
  } else if (DETAIL ("menubar")) {
    if (!style->bg_pixmap[state_type] || 
	gdk_window_get_type (window) == GDK_WINDOW_PIXMAP) {
      if (area)
        gdk_gc_set_clip_rectangle (style->bg_gc[state_type], area);

      gdk_draw_rectangle (window, style->bg_gc[state_type], TRUE,
		  	  x, y, width, height);
      if (area)
        gdk_gc_set_clip_rectangle (style->bg_gc[state_type], NULL);
    } else
      gtk_style_apply_default_background (style, window,
                                          widget && !GTK_WIDGET_NO_WINDOW (widget),
                                          state_type, area, x, y, width, height);
  } else if (DETAIL("notebook")) {
    GtkNotebook *notebook;
    GdkWindow *window_a, *window_b, *window_c;
    int a_x, a_y, b_x, b_y, c_x, c_y;
    int dx, dy;

    draw_box (style, window, state_type, shadow_type, area, widget,
	      NULL, x, y, width, height);
    parent_style = gtk_widget_get_style(GTK_WIDGET(widget->parent));

    if (!parent_style)
      parent_style = style;

    if (GTK_IS_NOTEBOOK (widget))
      notebook = GTK_NOTEBOOK (widget);
    else
      return;

    /* The notebook shadow is draw in three parts, on either the
       notebook window or the parent window depending on the
       orientation of the tabs. */

    gdk_window_get_position (window, &dx, &dy);

    window_a = window_b = window_c = widget->parent->window;
    a_x = b_x = c_x = x + dx;
    a_y = b_y = c_y = y + dy; 
 
    if (notebook->show_tabs) {
      switch (notebook->tab_pos) {
      case GTK_POS_LEFT:
      case GTK_POS_TOP:
	break;
      case GTK_POS_RIGHT:
	window_a = window;
	a_x = x;
	a_y = y;
	break;
      case GTK_POS_BOTTOM:
	break;
	window_c = window;
	c_x = x;
	c_y = y;
	break;
      }
    }

    mult (a, parent_style->bg[state_type], 0.814814815, style->colormap);
    gdk_gc_set_foreground (style->white_gc, &a);
    gdk_draw_line (window_a, style->white_gc, 
		   a_x+width, a_y+2, a_x+width, a_y+height-1);
    gdk_draw_line (window_c, style->white_gc, 
		   c_x+2, c_y+height, c_x+width-1, c_y+height);
    mult (a, parent_style->bg[state_type], 0.962962963, style->colormap);
    gdk_gc_set_foreground (style->white_gc, &a);
    gdk_draw_line (window_a, style->white_gc, 
		   a_x+width+1, a_y+2, a_x+width+1, a_y+height-1);
    gdk_draw_line (window_c, style->white_gc, 
		   c_x+2, c_y+height+1, c_x+width-1, c_y+height+1);
    mult (a, parent_style->bg[state_type], 0.9675925930, style->colormap);
    gdk_gc_set_foreground (style->white_gc, &a);
    gdk_draw_point (window_a, style->white_gc, a_x + width, a_y);
    gdk_draw_point (window_a, style->white_gc, a_x + width+1, a_y+1);
    gdk_draw_point (window_b, style->white_gc, b_x + width+1, b_y+height);
    gdk_draw_point (window_b, style->white_gc, b_x + width, b_y+height+1);
    gdk_draw_point (window_c, style->white_gc, c_x + 1, c_y+height+1);
    gdk_draw_point (window_c, style->white_gc, c_x, c_y+height);
    mult (a, parent_style->bg[state_type], 0.842592593, style->colormap);
    gdk_gc_set_foreground (style->white_gc, &a);
    gdk_draw_point (window_a, style->white_gc, a_x + width, a_y+1);
    gdk_draw_point (window_b, style->white_gc, b_x + width, b_y+height);
    gdk_draw_point (window_c, style->white_gc, c_x + 1, c_y+height);
    mult (a, parent_style->bg[state_type], 0.990740741, style->colormap);
    gdk_gc_set_foreground (style->white_gc, &a);
    gdk_draw_point (window_a, style->white_gc, a_x + width+1, a_y);
    gdk_draw_point (window_b, style->white_gc, b_x + width+1, b_y+height+1);
    gdk_draw_point (window_c, style->white_gc, c_x, c_y+height+1);
  } else if (DETAIL("hruler") || DETAIL ("vruler")) {
    draw_box (style, window, state_type, GTK_SHADOW_NONE, area, widget,
	      NULL, x, y, width, height);   
  } else {
    if (!style->bg_pixmap[state_type] || 
	gdk_window_get_type (window) == GDK_WINDOW_PIXMAP) {
      if (area)
        gdk_gc_set_clip_rectangle (style->bg_gc[state_type], area);

      gdk_draw_rectangle (window, style->bg_gc[state_type], TRUE,
		  	  x, y, width, height);
      if (area)
        gdk_gc_set_clip_rectangle (style->bg_gc[state_type], NULL);
    } else
      gtk_style_apply_default_background (style, window,
                                          widget && !GTK_WIDGET_NO_WINDOW (widget),
                                          state_type, area, x, y, width, height);
  
    gtk_paint_shadow (style, window, state_type, shadow_type,
                      area, widget, detail, x, y, width, height);
  }

  gdk_gc_set_foreground (style->white_gc, &style->white);
}

/**************************************************************************/

/**************************************************************************/
static void
draw_flat_box(GtkStyle * style,
	      GdkWindow * window,
	      GtkStateType state_type,
	      GtkShadowType shadow_type,
	      GdkRectangle * area,
	      GtkWidget * widget,
	      gchar * detail,
	      gint x,
	      gint y,
	      gint width,
	      gint height)
{
  GdkGC *gc1;
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1)) gdk_window_get_size(window, &width, &height);
  else if (width == -1)                gdk_window_get_size(window, &width, NULL);
  else if (height == -1)               gdk_window_get_size(window, NULL, &height);
  
  if (detail) {
    if (!strcmp ("text", detail) && state_type == GTK_STATE_SELECTED)
      gc1 = style->bg_gc[GTK_STATE_SELECTED];
    else if (!strcmp ("viewportbin", detail))
      gc1 = style->bg_gc[GTK_STATE_NORMAL];
    else if (!strcmp ("entry_bg", detail))
      gc1 = style->base_gc[state_type];
    else
      gc1 = style->bg_gc[state_type];
  } else
    gc1 = style->bg_gc[state_type];

  if (area) {
    gdk_gc_set_clip_rectangle(gc1, area);
    gdk_gc_set_clip_rectangle(style->fg_gc[state_type], area);
  }

  if (!style->bg_pixmap[state_type] || gc1 != style->bg_gc[state_type] ||
      gdk_window_get_type (window) == GDK_WINDOW_PIXMAP) {
    gdk_draw_rectangle(window, gc1, TRUE, x, y, width, height);
    if (DETAIL("tooltip")) {
      gdk_draw_rectangle(window, style->fg_gc[state_type], FALSE, x, y, width - 1, height - 1);
    }
  } else {
    gtk_style_apply_default_pixmap(style, window, state_type, area, x, y, width, height);
    if (DETAIL("tooltip")) {
      gdk_draw_rectangle(window, style->fg_gc[state_type], FALSE, x, y, width - 1, height - 1);
    }
  }

  if (area) {
    gdk_gc_set_clip_rectangle(gc1, NULL);
    gdk_gc_set_clip_rectangle(style->fg_gc[state_type], NULL);
  }
}
/**************************************************************************/
static void
draw_check(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  /* Fixed size only */

  if (area) {
    gdk_gc_set_clip_rectangle(style->bg_gc[state_type], area);
    gdk_gc_set_clip_rectangle(style->fg_gc[state_type], area);
  }

  x += width / 2 - 7;
  y += height / 2 - 7;
  /* Draw box */
/*
  if (shadow_type == GTK_SHADOW_IN && state_type == GTK_STATE_ACTIVE)
    state_type = GTK_STATE_NORMAL;
*/
  draw_button(style, window, state_type, shadow_type,
	      area, widget, detail, x, y, 14, 14);  

  if (shadow_type == GTK_SHADOW_IN) {
    gdk_gc_set_line_attributes (style->fg_gc[state_type],
                                2, GDK_LINE_SOLID, GDK_CAP_BUTT,
				GDK_JOIN_MITER);
    gdk_draw_line(window, style->fg_gc[state_type], x+3, y+3, x+10, y+10);
    gdk_draw_line(window, style->fg_gc[state_type], x+3, y+10, x+10, y+3);
    gdk_gc_set_line_attributes (style->fg_gc[state_type],
                                0, GDK_LINE_SOLID, GDK_CAP_BUTT,
				GDK_JOIN_MITER);
  }

  if (area) {
    gdk_gc_set_clip_rectangle(style->bg_gc[state_type], NULL);
    gdk_gc_set_clip_rectangle(style->fg_gc[state_type], NULL);
  }
}
/**************************************************************************/

/* Hopefully what you see below will be gone with future versions of gdk. */

/* These images were made in the gimp exported to C and striped for
   their alpha channels. */

/* 14 x 14 square - 14 x 14 circle */
unsigned char background0[] = 
  "\377\377\377\342w,\5\5"
  ",w\342\377\377\377\377\377"
  "\233\5\0\0\0\0\0\0"
  "\5\233\377\377\377\233\0\0"
  "\0\0\0\0\0\0\0\0\233"
  "\377\342\5\0\0\0\0\0"
  "\0\0\0\0\0\5\342w\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0w,\0\0\0\0\0\0"
  "\0\0\0\0\0\0,\5\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\5\5\0\0\0\0\0"
  "\0\0\0\0\0\0\0\5,"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0,w\0\0\0\0"
  "\0\0\0\0\0\0\0\0w\342"
  "\5\0\0\0\0\0\0\0\0"
  "\0\0\5\342\377\233\0\0"
  "\0\0\0\0\0\0\0\0\233"
  "\377\377\377\233\5\0\0\0"
  "\0\0\0\5\233\377\377\377"
  "\377\377\342w,\5\5,"
  "w\342\377\377\377";

/* 14 x 14 circle - 12 x 12 circle */
unsigned char circle0[] = 
  "\0\0\0\35\210\323\372\372"
  "\323\210\35\0\0\0\0\0"
  "d\372\213""4\6\6""4\213"
  "\372d\0\0\0d\335""4\0"
  "\0\0\0\0\0""4\335d\0"
  "\35\372""4\0\0\0\0\0"
  "\0\0\0""4\372\35\210\213"
  "\0\0\0\0\0\0\0\0\0"
  "\0\213\210\323""4\0\0\0\0"
  "\0\0\0\0\0\0""4\323"
  "\372\6\0\0\0\0\0\0\0"
  "\0\0\0\6\372\372\6\0\0"
  "\0\0\0\0\0\0\0\0\6"
  "\372\323""4\0\0\0\0\0"
  "\0\0\0\0\0""4\323\210\213"
  "\0\0\0\0\0\0\0\0\0"
  "\0\213\210\35\372""4\0\0"
  "\0\0\0\0\0\0""4\372"
  "\35\0d\335""4\0\0\0\0"
  "\0\0""4\335d\0\0\0d"
  "\372\213""4\6\6""4\213\372"
  "d\0\0\0\0\0\35\210\323"
  "\372\372\323\210\35\0\0\0";

/* 12 x 12 circle - 10 x 10 circle */
unsigned char circle1[] = 
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "t\313\371\371\313t\0\0"
  "\0\0\0\0\"\313\246?\10"
  "\10?\246\313\"\0\0\0"
  "\0\313s\0\0\0\0\0\0"
  "s\313\0\0\0t\246\0\0"
  "\0\0\0\0\0\0\246t\0"
  "\0\313?\0\0\0\0\0\0"
  "\0\0?\313\0\0\371\10"
  "\0\0\0\0\0\0\0\0\10"
  "\371\0\0\371\10\0\0\0\0"
  "\0\0\0\0\10\371\0\0\313"
  "?\0\0\0\0\0\0\0\0"
  "?\313\0\0t\246\0\0\0"
  "\0\0\0\0\0\246t\0\0"
  "\0\313s\0\0\0\0\0\0"
  "s\313\0\0\0\0\"\313"
  "\246?\10\10?\246\313\"\0"
  "\0\0\0\0\0t\313\371"
  "\371\313t\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0";

/* 10 x 10 circle - 8 x 8 circle */
unsigned char circle2[] = 
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0Y\300\367"
  "\367\300Y\0\0\0\0\0\0"
  "\0\214\315O\12\12O\315"
  "\214\0\0\0\0\0Y\315\12"
  "\0\0\0\0\12\315Y\0\0"
  "\0\0\300O\0\0\0\0\0"
  "\0O\300\0\0\0\0\367"
  "\12\0\0\0\0\0\0\12\367"
  "\0\0\0\0\367\12\0\0\0"
  "\0\0\0\12\367\0\0\0"
  "\0\300O\0\0\0\0\0\0"
  "O\300\0\0\0\0Y\315\12"
  "\0\0\0\0\12\315Y\0\0"
  "\0\0\0\214\315O\12\12"
  "O\315\214\0\0\0\0\0\0"
  "\0Y\300\367\367\300Y\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0";

/* 8 x 8 circle - 6 x 6 circle */
unsigned char circle3[] =
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0""2\260\365\365\260""2"
  "\0\0\0\0\0\0\0""2\365"
  "j\15\15j\365""2\0\0\0"
  "\0\0\0\260j\0\0\0\0"
  "j\260\0\0\0\0\0\0\365"
  "\15\0\0\0\0\15\365\0"
  "\0\0\0\0\0\365\15\0\0"
  "\0\0\15\365\0\0\0\0\0"
  "\0\260j\0\0\0\0j\260"
  "\0\0\0\0\0\0""2\365"
  "j\15\15j\365""2\0\0\0"
  "\0\0\0\0""2\260\365\365"
  "\260""2\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0";

/* 8 x 8 circle */
unsigned char circle4[] =
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0""2\260\365\365\260""2"
  "\0\0\0\0\0\0\0""2\365"
  "\377\377\377\377\365""2\0\0"
  "\0\0\0\0\260\377\377\377"
  "\377\377\377\260\0\0\0\0"
  "\0\0\365\377\377\377\377\377"
  "\377\365\0\0\0\0\0\0"
  "\365\377\377\377\377\377\377"
  "\365\0\0\0\0\0\0\260\377"
  "\377\377\377\377\377\260\0"
  "\0\0\0\0\0""2\365\377\377"
  "\377\377\365""2\0\0\0\0"
  "\0\0\0""2\260\365\365\260"
  """2\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0";

/* 6 x 6 circle */
unsigned char circle5[] =
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\225"
  "\362\362\225\0\0\0\0\0\0"
  "\0\0\0\225\377\377\377\377"
  "\225\0\0\0\0\0\0\0"
  "\0\362\377\377\377\377\362\0"
  "\0\0\0\0\0\0\0\362\377"
  "\377\377\377\362\0\0\0\0"
  "\0\0\0\0\225\377\377\377"
  "\377\225\0\0\0\0\0\0"
  "\0\0\0\225\362\362\225\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0\0\0\0"
  "\0\0\0\0\0\0";

typedef struct {
  double ax, ay, bx, by, vx, vy;
  double length;
  GdkColor c0, c1;
}LinearGradiant;

#define linear_grad(c, x, y, g)							\
do {										\
  double vx, vy;								\
  double mag;									\
  										\
  vx = (x) - g.ax;								\
  vy = (y) - g.ay;								\
										\
  mag = (g.vx * vx + g.vy * vy) / g.length;					\
										\
  if (mag <= 0.0)								\
    c = g.c0;									\
  else if (mag >= g.length)							\
    c = g.c1;									\
  else {									\
    mag /= g.length;								\
    c.red = (unsigned short)(g.c0.red * (1.0 - mag) + g.c1.red * mag);		\
    c.green = (unsigned short)(g.c0.green * (1.0 - mag) + g.c1.green * mag); 	\
    c.blue = (unsigned short)(g.c0.blue * (1.0 - mag) + g.c1.blue * mag);	\
  }										\
}while (0)

static void
draw_option(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height)
{
  GdkColor c, g[3];
  GtkStyle *parent;
  LinearGradiant g0, g1, g2;
  GdkImage *image;
  int i, j, value;

  if (state_type != GTK_STATE_PRELIGHT)
    state_type = GTK_STATE_NORMAL;

  if (state_type == GTK_STATE_NORMAL && widget->parent)
    parent = widget->parent->style;
  else
    parent = widget->style;

  g0.ax = 7.0 + 7.0 * cos (M_PI * 3.0 / 4.0);
  g0.ay = 7.0 - 7.0 * sin (M_PI * 3.0 / 4.0);
  g0.bx = 7.0 + 7.0 * cos (M_PI * 7.0 / 4.0);
  g0.by = 7.0 - 7.0 * sin (M_PI * 7.0 / 4.0);
  g0.vx = g0.bx - g0.ax;
  g0.vy = g0.by - g0.ay;
  g0.length = sqrt (g0.vx * g0.vx + g0.vy * g0.vy);
  mult (g0.c0, parent->bg[state_type], 0.699074074, style->colormap);
  mult (g0.c1, parent->bg[state_type], 1.111111111, style->colormap);
  g1.ax = 7.0 + 5.0 * cos (M_PI * 3.0 / 4.0);
  g1.ay = 7.0 - 5.0 * sin (M_PI * 3.0 / 4.0);
  g1.bx = 7.0 + 5.0 * cos (M_PI * 7.0 / 4.0);
  g1.by = 7.0 - 5.0 * sin (M_PI * 7.0 / 4.0);
  g1.vx = g1.bx - g1.ax;
  g1.vy = g1.by - g1.ay;
  g1.length = sqrt (g1.vx * g1.vx + g1.vy * g1.vy);
  mult (g1.c0, style->bg[state_type], 1.111111111, style->colormap);
  mult (g1.c1, style->bg[state_type], 0.699074074, style->colormap);
  g2.ax = 7.0 + 4.0;
  g2.ay = 7.0 - 4.0;
  g2.bx = 7.0 + 4.0;
  g2.by = 7.0 + 4.0;
  g2.vx = g2.bx - g2.ax;
  g2.vy = g2.by - g2.ay;
  g2.length = sqrt (g2.vx * g2.vx + g2.vy * g2.vy);
  mult (g2.c0, style->bg[state_type], 1.027777778, style->colormap);
  mult (g2.c1, style->bg[state_type], 0.861111111, style->colormap);

  image = gdk_image_new (GDK_IMAGE_FASTEST,
			   gdk_colormap_get_visual (style->colormap), 14, 14);

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, area);
  if (shadow_type == GTK_SHADOW_IN) {
    for (i = 0; i < 14; i++) {
      for (j = 0; j < 14; j++) {
	linear_grad (g[0], j+0.5, i+0.5, g0);
	linear_grad (g[1], j+0.5, i+0.5, g1);
	linear_grad (g[2], j+0.5, i+0.5, g2);
 
	value = parent->bg[state_type].red * background0[i*14+j] * 257 / 0xffff
	  + g[0].red * circle0[i*14+j] * 257 / 0xffff
	  + parent->bg[state_type].red * 0.285493827 * circle1[i*14+j] * 257 / 0xffff
	  + g[1].red * circle2[i*14+j] * 257 / 0xffff
	  + g[2].red * circle3[i*14+j] * 257 / 0xffff
	  + style->fg[state_type].red * circle5[i*14+j] * 257 / 0xffff;
	c.red = value > 0xffff? 0xffff : value;
	value = parent->bg[state_type].green * background0[i*14+j] * 257 / 0xffff
	  + g[0].green * circle0[i*14+j] * 257 / 0xffff
	  + parent->bg[state_type].green * 0.285493827 * circle1[i*14+j] * 257 / 0xffff
	  + g[1].green * circle2[i*14+j] * 257 / 0xffff
	  + g[2].green * circle3[i*14+j] * 257 / 0xffff
	  + style->fg[state_type].green * circle5[i*14+j] * 257 / 0xffff;
	c.green = value > 0xffff? 0xffff : value;
	value = parent->bg[state_type].blue * background0[i*14+j] * 257 / 0xffff
	  + g[0].blue * circle0[i*14+j] * 257 / 0xffff
	  + parent->bg[state_type].blue * 0.285493827 * circle1[i*14+j] * 257 / 0xffff
	  + g[1].blue * circle2[i*14+j] * 257 / 0xffff
	  + g[2].blue * circle3[i*14+j] * 257 / 0xffff
	  + style->fg[state_type].blue * circle5[i*14+j] * 257 / 0xffff;
	c.blue = value > 0xffff? 0xffff : value;
	gdk_color_alloc (style->colormap, &c);

	gdk_image_put_pixel (image, j, i, c.pixel);
      }
    }
  } else {
    for (i = 0; i < 14; i++) {
      for (j = 0; j < 14; j++) {
	linear_grad (g[0], j+0.5, i+0.5, g0);
	linear_grad (g[1], j+0.5, i+0.5, g1);
	linear_grad (g[2], j+0.5, i+0.5, g2);
 
	value = parent->bg[state_type].red * background0[i*14+j] * 257 / 0xffff
	  + g[0].red * circle0[i*14+j] * 257 / 0xffff
	  + parent->bg[state_type].red * 0.285493827 * circle1[i*14+j] * 257 / 0xffff
	  + g[1].red * circle2[i*14+j] * 257 / 0xffff
	  + g[2].red * circle4[i*14+j] * 257 / 0xffff;
	c.red = value > 0xffff? 0xffff : value;
	value = parent->bg[state_type].green * background0[i*14+j] * 257 / 0xffff
	  + g[0].green * circle0[i*14+j] * 257 / 0xffff
	  + parent->bg[state_type].green * 0.285493827 * circle1[i*14+j] * 257 / 0xffff
	  + g[1].green * circle2[i*14+j] * 257 / 0xffff
	  + g[2].green * circle4[i*14+j] * 257 / 0xffff;
	c.green = value > 0xffff? 0xffff : value;
	value = parent->bg[state_type].blue * background0[i*14+j] * 257 / 0xffff
	  + g[0].blue * circle0[i*14+j] * 257 / 0xffff
	  + parent->bg[state_type].blue * 0.285493827 * circle1[i*14+j] * 257 / 0xffff 
	  + g[1].blue * circle2[i*14+j] * 257 / 0xffff
	  + g[2].blue * circle4[i*14+j] * 257 / 0xffff;
	c.blue = value > 0xffff? 0xffff : value;
	gdk_color_alloc (style->colormap, &c);

	gdk_image_put_pixel (image, j, i, c.pixel);
      }
    }
  }

  gdk_draw_image (window, style->white_gc, image,
                  0, 0, x + width / 2 - 7, y + height / 2 - 7, 14, 14);

  gdk_image_destroy (image);

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, NULL);
}
/**************************************************************************/
static void
draw_cross(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
}

/**************************************************************************/
static void
draw_ramp(GtkStyle * style,
	  GdkWindow * window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle * area,
	  GtkWidget * widget,
	  gchar * detail,
	  GtkArrowType arrow_type,
	  gint x,
	  gint y,
	  gint width,
	  gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
}

/**************************************************************************/
static void
draw_tab(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GtkShadowType shadow_type,
	 GdkRectangle * area,
	 GtkWidget * widget,
	 gchar * detail,
	 gint x,
	 gint y,
	 gint width,
	 gint height)
{
  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  gtk_paint_arrow (style, window, state_type, shadow_type, area,
                   widget, "arrow", GTK_ARROW_UP, TRUE, x, y-height/2, width, height);
  gtk_paint_arrow (style, window, state_type, shadow_type, area,
                   widget, "arrow", GTK_ARROW_DOWN, TRUE, x, y+height/2, width, height);
}

/**************************************************************************/
static void
draw_shadow_gap(GtkStyle * style,
		GdkWindow * window,
		GtkStateType state_type,
		GtkShadowType shadow_type,
		GdkRectangle * area,
		GtkWidget * widget,
		gchar * detail,
		gint x,
		gint y,
		gint width,
		gint height,
		GtkPositionType gap_side,
		gint gap_x,
		gint gap_width)
{
  GdkRectangle        rect;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  gtk_paint_shadow(style, window, state_type, shadow_type, area, widget, detail,
		   x, y, width, height);

  switch (gap_side)
    {
    case GTK_POS_TOP:
      rect.x = x + gap_x;
      rect.y = y;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_BOTTOM:
      rect.x = x + gap_x;
      rect.y = y + height - 2;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_LEFT:
      rect.x = x;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    case GTK_POS_RIGHT:
      rect.x = x + width - 2;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
  }

  gtk_style_apply_default_pixmap(style, window, state_type, area,
				 rect.x, rect.y, rect.width, rect.height);
}
/**************************************************************************/
static void
draw_box_gap(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     gint x,
	     gint y,
	     gint width,
	     gint height,
	     GtkPositionType gap_side,
	     gint gap_x,
	     gint gap_width)
{
  GdkRectangle        rect;
  GdkColor c;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
		x, y, width, height);

  if (gap_x < 2) {
    gap_width -= 2 - gap_x;
    gap_x = 2;
    switch (gap_side) {
    case GTK_POS_TOP:
      mult (c, style->bg[state_type], 1.111111111, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_point (window, style->white_gc, x + 1, y);
      break;
    case GTK_POS_BOTTOM:
      mult (c, style->bg[state_type], 1.111111111, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_point (window, style->white_gc, x + 1, y+height-2);
      gdk_draw_point (window, style->white_gc, x + 1, y+height-1);
      break;
    case GTK_POS_LEFT:
      mult (c, style->bg[state_type], 1.111111111, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_point (window, style->white_gc, x + width - 2, y+1);
      break;
    case GTK_POS_RIGHT:
      mult (c, style->bg[state_type], 1.111111111, style->colormap);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_point (window, style->white_gc, x + width - 2, y+1);
      break;    
    }   
  } 

  switch (gap_side) {
    case GTK_POS_TOP:
      rect.x = x + gap_x;
      rect.y = y;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_BOTTOM:
      rect.x = x + gap_x;
      rect.y = y + height - 2;
      rect.width = gap_width;
      rect.height = 2;
      break;
    case GTK_POS_LEFT:
      rect.x = x;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
    case GTK_POS_RIGHT:
      rect.x = x + width - 2;
      rect.y = y + gap_x;
      rect.width = 2;
      rect.height = gap_width;
      break;
  }

  gtk_style_apply_default_pixmap(style, window, state_type, area,
				 rect.x, rect.y, rect.width, rect.height);

}
/**************************************************************************/
static void
draw_extension(GtkStyle * style,
	       GdkWindow * window,
	       GtkStateType state_type,
	       GtkShadowType shadow_type,
	       GdkRectangle * area,
	       GtkWidget * widget,
	       gchar * detail,
	       gint x,
	       gint y,
	       gint width,
	       gint height,
	       GtkPositionType gap_side)
{
  GdkColor c;
  GdkRectangle rect;
  int tx, ty, twidth, theight;

  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
  
  if (width == -1 && height == -1)
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);
  
  if (shadow_type == GTK_SHADOW_NONE)
    return;

  rect.x = tx = x;
  rect.y = ty = y;
  rect.width = twidth = width;
  rect.height = theight = height;
  
  switch (gap_side) {
  case GTK_POS_TOP:
    ty -= 2;
  case GTK_POS_BOTTOM:
    theight += 2;
    break;
  case GTK_POS_LEFT:
    tx -= 2;
  case GTK_POS_RIGHT:
    twidth += 2;
    break;
  }
  
  gtk_style_apply_default_background (style, window,
				      widget && !GTK_WIDGET_NO_WINDOW (widget),
				      state_type, area,
				      tx + 2, 
				      ty + 2, 
				      twidth - 4, 
				      theight - 4);
  draw_shadow(style, window, state_type, shadow_type,
	      &rect, widget, NULL, tx, ty, twidth, theight);

  /* tab shadow */
  if (GTK_IS_NOTEBOOK (widget) && state_type == GTK_STATE_NORMAL) {
    GtkStyle *parent_style;
    gboolean last_tab;
    GtkNotebook *notebook = GTK_NOTEBOOK (widget);
    GList *last = g_list_last (notebook->children);
    GdkColor n[5], d[6];
    int dx, dy;

    if (notebook->cur_page == last->data) {
      last_tab = TRUE;
    } else {
      last_tab = FALSE;
    }

    parent_style = gtk_widget_get_style(GTK_WIDGET(widget->parent));

    if (!parent_style)
      parent_style = style;

    mult (n[0], parent_style->bg[state_type], 0.814814815, style->colormap);
    mult (n[1], parent_style->bg[state_type], 0.962962963, style->colormap);
    mult (n[2], parent_style->bg[state_type], 0.9675925930, style->colormap);
    mult (n[3], parent_style->bg[state_type], 0.842592593, style->colormap);
    mult (n[4], parent_style->bg[state_type], 0.990740741, style->colormap);

    if (!last_tab) {
      mult (d[0], style->bg[GTK_STATE_ACTIVE], 0.814814815, style->colormap);
      mult (d[1], style->bg[GTK_STATE_ACTIVE], 0.962962963, style->colormap);
      mult (d[2], style->bg[GTK_STATE_ACTIVE], 0.814814815 * 1.111111111, style->colormap);
      mult (d[3], style->bg[GTK_STATE_ACTIVE], 0.962962963 * 1.111111111, style->colormap);
      mult (d[4], style->bg[GTK_STATE_ACTIVE], 0.814814815 * 0.285493827, style->colormap);
      mult (d[5], style->bg[GTK_STATE_ACTIVE], 0.962962963 * 0.285493827, style->colormap);     
    } else {
      gdk_window_get_position (window, &dx, &dy);
    }

    /* The shadow is five parts. */
    /* 1 - top right, present in BOTTOM and LEFT. */
    if (gap_side == GTK_POS_BOTTOM || gap_side == GTK_POS_LEFT) {
      if (last_tab) {
	gdk_gc_set_foreground (style->white_gc, &n[2]);
	gdk_draw_point (widget->parent->window, style->white_gc, x +dx+ width, y+dy);
	gdk_draw_point (widget->parent->window, style->white_gc, x +dx+ width + 1, y+dy+1);
	gdk_gc_set_foreground (style->white_gc, &n[3]); 
	gdk_draw_point (widget->parent->window, style->white_gc, x +dx+ width, y+dy+1);
	gdk_gc_set_foreground (style->white_gc, &n[4]); 
	gdk_draw_point (widget->parent->window, style->white_gc, x +dx+ width + 1, y+dy); 
      } else {
	gdk_gc_set_foreground (style->white_gc, &n[2]);
	gdk_draw_point (window, style->white_gc, x + width, y);
	gdk_draw_point (window, style->white_gc, x + width + 1, y+1);
	gdk_gc_set_foreground (style->white_gc, &n[3]); 
	gdk_draw_point (window, style->white_gc, x + width, y+1);
	gdk_gc_set_foreground (style->white_gc, &n[4]); 
	gdk_draw_point (window, style->white_gc, x + width + 1, y);
      }
    }

    /* 2 - right side, present in BOTTOM, TOP, and LEFT */
    if (gap_side != GTK_POS_RIGHT) {
      if (last_tab) {
	gdk_gc_set_foreground (style->white_gc, &n[0]);
	if (gap_side == GTK_POS_TOP)
	  gdk_draw_line (widget->parent->window, style->white_gc, 
			 x+dx+width, y+dy+1, x+dx+width, y+dy+height-1);
	else
	  gdk_draw_line (widget->parent->window, style->white_gc, 
			 x+dx+width, y+dy+2, x+dx+width, y+dy+height-1);
	gdk_gc_set_foreground (style->white_gc, &n[1]);
	gdk_draw_line (widget->parent->window, style->white_gc, 
		       x+dx+width+1, y+dy+2, x+dx+width+1, y+dy+height-1);
      } else {
	switch (gap_side) {
	case GTK_POS_LEFT:
	  gdk_gc_set_foreground (style->white_gc, &d[0]);
	  gdk_draw_line (window, style->white_gc, 
			 x+width, y+2, x+width, y+height-1);
	  gdk_gc_set_foreground (style->white_gc, &d[1]);
	  gdk_draw_line (window, style->white_gc, 
			 x+width+1, y+2, x+width+1, y+height-1);
	  break;
	case GTK_POS_BOTTOM:
	  gdk_gc_set_foreground (style->white_gc, &d[0]);
	  gdk_draw_line (window, style->white_gc, 
			 x+width, y+4, x+width, y+height-1);
	  gdk_gc_set_foreground (style->white_gc, &d[1]);
	  gdk_draw_line (window, style->white_gc, 
			 x+width+1, y+4, x+width+1, y+height-1);
	  gdk_gc_set_foreground (style->white_gc, &d[2]);
	  gdk_draw_point (window, style->white_gc,  x+width, y+3);
	  gdk_gc_set_foreground (style->white_gc, &d[3]);
	  gdk_draw_point (window, style->white_gc, x+width+1, y+3);
	  gdk_gc_set_foreground (style->white_gc, &d[4]);
	  gdk_draw_point (window, style->white_gc,  x+width, y+2);
	  gdk_gc_set_foreground (style->white_gc, &d[5]);
	  gdk_draw_point (window, style->white_gc, x+width+1, y+2);
	  break;
	case GTK_POS_TOP:
	  gdk_gc_set_foreground (style->white_gc, &d[0]);
	  gdk_draw_line (window, style->white_gc, 
			 x+width, y, x+width, y+height-5);
	  gdk_gc_set_foreground (style->white_gc, &d[1]);
	  gdk_draw_line (window, style->white_gc, 
			 x+width+1, y, x+width+1, y+height-5);
	  gdk_gc_set_foreground (style->white_gc, &d[2]);
	  gdk_draw_point (window, style->white_gc,  x+width, y+height-4);
	  gdk_gc_set_foreground (style->white_gc, &d[3]);
	  gdk_draw_point (window, style->white_gc, x+width+1, y+height-4);
	  gdk_gc_set_foreground (style->white_gc, &d[4]);
	  gdk_draw_point (window, style->white_gc,  x+width, y+height-3);
	  gdk_gc_set_foreground (style->white_gc, &d[5]);
	  gdk_draw_point (window, style->white_gc, x+width+1, y+height-3);
	  break;
	}
      }
    }

    /* 3 - bottom right, present in LEFT and TOP */
    if (gap_side == GTK_POS_LEFT || gap_side == GTK_POS_TOP) {
      gdk_gc_set_foreground (style->white_gc, &n[2]);
      gdk_draw_point (window, style->white_gc, x + width + 1, y+height);
      gdk_draw_point (window, style->white_gc, x + width, y+height+1);
      gdk_gc_set_foreground (style->white_gc, &n[3]);
      gdk_draw_point (window, style->white_gc, x + width, y+height);
      gdk_gc_set_foreground (style->white_gc, &n[4]);
      gdk_draw_point (window, style->white_gc, x + width + 1, y+height+1);
    }

    /* 4 - bottom, present in RIGHT, LEFT and TOP  */
    if (gap_side != GTK_POS_BOTTOM) {
      if (last_tab) {
	gdk_gc_set_foreground (style->white_gc, &n[0]);
	gdk_draw_line (window, style->white_gc, 
		       x+2, y+height, x+width-1, y+height);
	gdk_gc_set_foreground (style->white_gc, &n[1]);
	gdk_draw_line (window, style->white_gc, 
		       x+2, y+height+1, x+width-1, y+height+1);
      } else {
	gdk_gc_set_foreground (style->white_gc, &d[0]);
	gdk_draw_line (window, style->white_gc, 
		       x+2, y+height, x+width-1, y+height);
	gdk_gc_set_foreground (style->white_gc, &d[1]);
	gdk_draw_line (window, style->white_gc, 
		       x+2, y+height+1, x+width-1, y+height+1);
      }
    }

    /* 5 - bottom left, present in RIGHT and TOP */
    if (gap_side == GTK_POS_RIGHT || gap_side == GTK_POS_TOP) {
      gdk_gc_set_foreground (style->white_gc, &n[2]);
      gdk_draw_point (window, style->white_gc, x + 1, y+height+1);
      gdk_draw_point (window, style->white_gc, x, y+height);
      gdk_gc_set_foreground (style->white_gc, &n[3]);
      gdk_draw_point (window, style->white_gc, x + 1, y+height);
      gdk_gc_set_foreground (style->white_gc, &n[4]);
      gdk_draw_point (window, style->white_gc, x, y+height+1);
    }
  }

  if (area)
    gdk_gc_set_clip_rectangle (style->white_gc, NULL);
}
/**************************************************************************/
static void
draw_focus(GtkStyle * style,
	   GdkWindow * window,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
  GdkGC *focusgc;
  int state;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1)) gdk_window_get_size(window, &width, &height);
  else if (width == -1)                gdk_window_get_size(window, &width, NULL);
  else if (height == -1)               gdk_window_get_size(window, NULL, &height);

  state = GTK_WIDGET_STATE(widget);
  focusgc = style->fg_gc[state];

  if (area) {
    gdk_gc_set_clip_rectangle(focusgc, area);
  }

  if (DETAIL("button") || DETAIL("togglebutton") || 
      DETAIL("buttondefault")) {
    gdk_draw_rectangle(window, focusgc, FALSE, x+1, y+1, width-2, height-2);
  } else if (DETAIL("checkbutton") || DETAIL("radiobutton") || DETAIL ("label")) {
  } else {
    gdk_draw_rectangle(window, focusgc, FALSE, x, y, width-1, height-1);
  }

  if (area) {
    gdk_gc_set_clip_rectangle(focusgc, NULL);
  }
}
/**************************************************************************/
static void
draw_slider(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height,
	    GtkOrientation orientation)
{
  GdkPixmap *p;
  GdkColor a, b;

  if ((width == -1) && (height == -1)) gdk_window_get_size(window, &width, &height);
  else if (width == -1)                gdk_window_get_size(window, &width, NULL);
  else if (height == -1)               gdk_window_get_size(window, NULL, &height);

  p = gdk_pixmap_new (window, width, height, style->depth);

  gtk_paint_shadow (style, p, GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		    area, widget, NULL, x, y, width, height);

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, area);

  draw_concave_box (style, p, area, x+2, y+2, width-4, height-4,
		    &style->bg[state_type]);

  /* Reset Clip Region */
  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, NULL);

  /* Draw middle line */
  if (GTK_CHECK_TYPE(widget, gtk_hscale_get_type())) {
    draw_vline(style, p, state_type, area, widget, detail,
               y+2, y+height-3, x+width/2);
  } else {
    draw_hline(style, p, state_type, area, widget, detail,
                  x+2, x+width-4, y+height/2);
  }
	gdk_window_set_back_pixmap (window, p, 0);
      gdk_pixmap_unref (p);
    
      if (area) {
	gdk_window_clear_area (window, area->x, area->y,
			     area->width, area->height);
      } else
	gdk_window_clear (window);
}
/**************************************************************************/
static void
draw_handle(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height,
	    GtkOrientation orientation)
{
  GtkStyle *parent_style;
  GdkPixmap *p;
  GdkColor a, b;

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, area);

  parent_style = gtk_widget_get_style(GTK_WIDGET(widget->parent));
  gtk_paint_shadow (parent_style, window, GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		    area, widget, NULL, x, y, width, height);
  draw_concave_box (style, window, area, x+2, y+2, width-4, height-4,
		    &style->bg[state_type]);

  /* Reset Clip Region */
  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, NULL);
}



/* special drawing functions */
static void 
draw_button(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height)
{
  GtkStyle *parent_style;
  GdkPixmap *p;
  GdkColor a, b;
  
  if (DETAIL ("buttondefault")) {
    parent_style = gtk_widget_get_style(GTK_WIDGET(widget->parent));
    if (!parent_style)
      parent_style = style;
    gtk_paint_shadow (parent_style, window, GTK_STATE_NORMAL, GTK_SHADOW_IN,
		      area, widget, NULL, x, y, width, height);
    draw_concave_box (style, window, area, x+2, y+2, width-4, height-4,
		      &style->bg[state_type]);
  } else {
    /* draw outer box */
    if (widget && GTK_IS_BUTTON (widget)
	&& GTK_WIDGET_HAS_DEFAULT (widget)
	&& GTK_BUTTON (widget)->relief == GTK_RELIEF_NORMAL)
      parent_style = style;
    else
      parent_style = gtk_widget_get_style(GTK_WIDGET(widget->parent));
    if (!parent_style)
      parent_style = style;
    gtk_paint_shadow (parent_style, window, GTK_STATE_NORMAL, GTK_SHADOW_IN,
		      area, widget, NULL, x, y, width, height);
  
    /* draw inner box */
    draw_beveled_box (style, window, area, x+2, y+2, width-4, height-4,
		      &style->bg[state_type]);
  }


  /* Reset Clip Region */

}


void
draw_trough (GtkStyle * style,
	     GdkWindow * window,
	     GdkRectangle * area,
	     int x, int y, int width, int height,
	     GdkColor *c)
{
  GdkColor a, b;

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, area);

  a.red = (unsigned short)(c->red * 0.393518519);
  a.green = (unsigned short)(c->green * 0.393518519);
  a.blue = (unsigned short)(c->blue * 0.393518519);
  gdk_color_alloc (style->colormap, &a);
  gdk_gc_set_foreground (style->white_gc, &a);
  gdk_draw_line (window, style->white_gc, 
		 x, y, x+width-1, y);
  a.red = (unsigned short)(c->red * 0.439814815);
  a.green = (unsigned short)(c->green * 0.439814815);
  a.blue = (unsigned short)(c->blue * 0.439814815);
  gdk_color_alloc (style->colormap, &a);
  gdk_gc_set_foreground (style->white_gc, &a);
  gdk_draw_point (window, style->white_gc, x, y+1);
  a.red = (unsigned short)(c->red * 0.486111111);
  a.green = (unsigned short)(c->green * 0.486111111);
  a.blue = (unsigned short)(c->blue * 0.486111111);
  gdk_color_alloc (style->colormap, &a);
  gdk_gc_set_foreground (style->white_gc, &a);
  gdk_draw_line (window, style->white_gc, 
		 x+1, y+1, x+width-1, y+1);
  a.red = (unsigned short)(c->red * 0.50308642);
  a.green = (unsigned short)(c->green * 0.50308642);
  a.blue = (unsigned short)(c->blue * 0.50308642);
  b.red = (unsigned short)(c->red * 0.587962963);
  b.green = (unsigned short)(c->green * 0.587962963);
  b.blue = (unsigned short)(c->blue * 0.587962963);
  draw_vertical_gradient (style, window, area,
			  x, y+2, 1, height-2, &a, &b);
  a.red = (unsigned short)(c->red * 0.560185185);
  a.green = (unsigned short)(c->green * 0.560185185);
  a.blue = (unsigned short)(c->blue * 0.560185185);
  b.red = (unsigned short)(c->red * 0.657407407);
  b.green = (unsigned short)(c->green * 0.657407407);
  b.blue = (unsigned short)(c->blue * 0.657407407);
  draw_vertical_gradient (style, window, area,
			  x+1, y+2, width-1, height-2, &a, &b);
  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, NULL);
}


void
draw_concave_box (GtkStyle * style,
		  GdkWindow * window,
		  GdkRectangle * area,
		  int x, int y, int width, int height,
		  GdkColor *c)
{
  GdkColor a, b;

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, area);

  a.red = clamp ((unsigned int)(c->red * 0.861111111));
  a.green = clamp ((unsigned int)(c->green * 0.861111111));
  a.blue = clamp ((unsigned int)(c->blue * 0.861111111));
  b.red = clamp ((unsigned int)(c->red * 1.027777778));
  b.green = clamp ((unsigned int)(c->green * 1.027777778));
  b.blue = clamp ((unsigned int)(c->blue * 1.027777778));
  draw_vertical_gradient (style, window, area, x, y,
			  width, height, &a, &b);
  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, NULL);
}

void
draw_beveled_box (GtkStyle * style,
		  GdkWindow * window,
		  GdkRectangle * area,
		  int x, int y, int width, int height,
		  GdkColor *c)
{
  GdkColor a, b;

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, area);

  a.red = clamp ((unsigned int)(c->red * 1.092592593));
  a.green = clamp ((unsigned int)(c->green * 1.092592593));
  a.blue = clamp ((unsigned int)(c->blue * 1.092592593));
  gdk_color_alloc (style->colormap, &a);
  gdk_gc_set_foreground (style->white_gc, &a);
  gdk_draw_line (window, style->white_gc, x, y, x+width-2, y);
  a.red = clamp ((unsigned int)(c->red * .958333333));
  a.green = clamp ((unsigned int)(c->green * .958333333));
  a.blue = clamp ((unsigned int)(c->blue * .958333333));
  gdk_color_alloc (style->colormap, &a);
  gdk_gc_set_foreground (style->white_gc, &a);
  gdk_draw_point (window, style->white_gc, x+width-1, y);

  a.red = clamp ((unsigned int)(c->red * 1.087962963));
  a.green = clamp ((unsigned int)(c->green * 1.087962963));
  a.blue = clamp ((unsigned int)(c->blue * 1.087962963));
  b.red = clamp ((unsigned int)(c->red * 0.976851852));
  b.green = clamp ((unsigned int)(c->green * 0.976851852));
  b.blue = clamp ((unsigned int)(c->blue * 0.976851852));
  draw_vertical_gradient (style, window, area, x, y+1, 1, height-2, &a, &b);
  a.red = clamp ((unsigned int)(c->red * 1.027777778));
  a.green = clamp ((unsigned int)(c->green * 1.027777778));
  a.blue = clamp ((unsigned int)(c->blue * 1.027777778));
  b.red = clamp ((unsigned int)(c->red * 0.861111111));
  b.green = clamp ((unsigned int)(c->green * 0.861111111));
  b.blue = clamp ((unsigned int)(c->blue * 0.861111111));
/*
  draw_vertical_gradient (style, window, area, x+1, y+1,
			  width-2, height-2, &a, &b);
*/
  draw_cubic_vertical_gradient (style, window, area, x+1, y+1,
				width-2, height-2,
				&a, 0.061476, -0.091382, -0.132226, 1.0);
  a.red = clamp ((unsigned int)(c->red * 0.824074074));
  a.green = clamp ((unsigned int)(c->green * 0.824074074));
  a.blue = clamp ((unsigned int)(c->blue * 0.824074074));
  b.red = clamp ((unsigned int)(c->red * 0.685185185));
  b.green = clamp ((unsigned int)(c->green * 0.685185185));
  b.blue = clamp ((unsigned int)(c->blue * 0.685185185));
  draw_vertical_gradient (style, window, area, x+width-1, y+1,
			  1, height-2, &a, &b);

  a.red = clamp ((unsigned int)(c->red * 0.787037037));
  a.green = clamp ((unsigned int)(c->green * 0.787037037));
  a.blue = clamp ((unsigned int)(c->blue * 0.787037037));
  gdk_color_alloc (style->colormap, &a);
  gdk_gc_set_foreground (style->white_gc, &a);
  gdk_draw_point (window, style->white_gc, x, y+height-1);
  a.red = clamp ((unsigned int)(c->red * 0.675925926));
  a.green = clamp ((unsigned int)(c->green * 0.675925926));
  a.blue = clamp ((unsigned int)(c->blue * 0.675925926));
  gdk_color_alloc (style->colormap, &a);
  gdk_gc_set_foreground (style->white_gc, &a);
  gdk_draw_line (window, style->white_gc, x+1, y+height-1,
		 x+width-1, y+height-1);

  gdk_gc_set_foreground (style->white_gc, &style->white);

  if (area)
    gdk_gc_set_clip_rectangle(style->white_gc, NULL);
}


void
draw_vertical_gradient (GtkStyle * style,
			GdkWindow * window,
			GdkRectangle * area,
			int x, int y, int width, int height,
			GdkColor *a,
			GdkColor *b)
{
  GdkColor c;
  GdkVisual *visual = gdk_visual_get_system ();
  int dr, dg, db, l, ty, i, j, max;

  dr = b->red - a->red;
  dg = b->green - a->green;
  db = b->blue - a->blue;

  /* compute the maximum color variation */
  i = dr < 0 ? -dr : dr;
  i >>= (16 - visual->red_prec);
  j = dg < 0 ? -dg : dg;
  j >>= (16 - visual->green_prec);
  max = i > j ? i : j;
  i = db < 0 ? -db : db;
  i >>= (16 - visual->blue_prec);
  max = max > i ? max : i;

  if (height <= max + 1) {
    /* draw gradient with lines */
    for (l = 0; l < height; l++) {
      ty = y + l;
      c.red = clamp (a->red + dr * l / height);
      c.green = clamp (a->green + dg * l / height);
      c.blue = clamp (a->blue + db * l / height);
      gdk_color_alloc (style->colormap, &c);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_line (window, style->white_gc,
                     x, ty, x + width - 1, ty);
    }
  } else {
    int inc = height / (max + 1);

    /* aproximate gradient with rectangles */
    for (l = 0; l+inc <= height; l += inc) {
      c.red = clamp (a->red + dr * l / height);
      c.green = clamp (a->green + dg * l / height);
      c.blue = clamp (a->blue + db * l / height);
      gdk_color_alloc (style->colormap, &c);
      gdk_gc_set_foreground (style->white_gc, &c);
      gdk_draw_rectangle (window, style->white_gc, TRUE,
                          x, y + l, width, inc);
    }

    if (l != height) {
      gdk_color_alloc (style->colormap, b);
      gdk_gc_set_foreground (style->white_gc, b);
      gdk_draw_rectangle (window, style->white_gc, TRUE,
                          x, y + l, width,
			  height - l); 
    }
  }

  gdk_gc_set_foreground (style->white_gc, &style->white);
}

void
draw_cubic_vertical_gradient (GtkStyle * style,
			      GdkWindow * window,
			      GdkRectangle * area,
			      int x, int y, int width, int height,
			      GdkColor *base_color,
			      float a, float b, float c, float d)
{
  GdkColor color;
  int l, ty;
  float t = 0;
  float inc = 1.0 / (height - 1);

  for (l = 0; l < height; l++) {
    float v;
    ty = y + l;
    t = inc * l;
    v = a*t*t*t + b*t*t + c*t + d; 

    color.red = (unsigned short)(base_color->red * v); 
    color.green = (unsigned short)(base_color->green * v);
    color.blue = (unsigned short)(base_color->blue * v); 
    gdk_color_alloc (style->colormap, &color);
    gdk_gc_set_foreground (style->white_gc, &color);
    gdk_draw_line (window, style->white_gc,
		   x, ty, x + width - 1, ty);
  }

  gdk_gc_set_foreground (style->white_gc, &style->white);
}
